;; Generated package description from futur.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "futur" "1.1.0.20260227.113135" "Future/promise-based async library" 'nil :commit "d320357774b675034b35e99ea350f8e901bf0a15" :url "https://elpa.gnu.org/packages/futur.html" :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca") :keywords '("concurrency" "async" "promises" "futures"))
